#include "gsmhacking.h"

int find_animation_stuff(struct gsm *gsa) {
  unsigned short i;
  unsigned int pos;

  if (! gsa || ! gsa->fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  if (read_short(gsa->fd, &gsa->animation_unknown_before_framerate)) {
    ERROR("can't read unknown before framerate");
    return 2;
  }
  DEBUG("unknown before framerate: %u\n", gsa->animation_unknown_before_framerate);
  if (read_short(gsa->fd, &gsa->framerate)) {
    ERROR("can't read framerate");
    return 3;
  }
  DEBUG("framerate: %ufps\n", gsa->framerate);
  if (read_short(gsa->fd, &gsa->animation_unknown_after_framerate)) {
    ERROR("can't read unknown after framerate");
    return 4;
  }
  DEBUG("unknown before framerate: %u\n", gsa->animation_unknown_after_framerate);

  return 0;
}

struct frame **alloc_frames(unsigned short n) {
  return (struct frame **) alloc_lots(n, "frame", sizeof(struct frame *), (void *(*)()) alloc_frame);
}

struct frame *alloc_frame() {
  return (struct frame *) alloc_one("frame", sizeof(struct frame));
}

void cleanup_frames(unsigned short num_frames, struct frame **frames) {
  unsigned short i;
  for (i = 0; i < num_frames; i++) {
    cleanup_events(frames[i]->num_events, frames[i]->events);
  }
  free_structs("frame", num_frames, (void **) frames);
}

int find_frames(struct gsm *gsa) {
  struct frame *frame;
  int bone_id;
  struct bone *bone;
  unsigned short i;
  unsigned int pos;

  if (! gsa || ! gsa->fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  /* read number of frames */
  if (read_short(gsa->fd, &gsa->num_frames)) {
    ERROR("can't read number of frames");
    return 2;
  }
  if (! gsa->num_frames) {
    SAY("%s has no frames!\n", gsa->filename);
    return 3;
  }
  DEBUG("frames: %u\n", gsa->num_frames);

  /* allocate a big enough struct */
  gsa->frames = alloc_frames(gsa->num_frames);
  if (! gsa->num_frames) return 4;

  /* and allocate frames for each bone */
  if (! gsa->skeletons[0] || ! gsa->skeletons[0]->num_bones) {
    ERROR("no bones in gsa\n");
    return 5;
  }

  /* we need to allocate matrices and quaternions even for non-animated bones */
  for (i = 0; i < gsa->skeletons[0]->num_bones; i++) {
    bone = gsa->skeletons[0]->bones[i];
    bone->orientations = alloc_quaternions(gsa->num_frames);
    if (! bone->orientations) {
      ERROR("out of memory for bone %u orientations", i);
      return 6;
    }
    bone->locals = alloc_matrices(gsa->num_frames);
    if (! bone->locals) {
      ERROR("out of memory for bone %u local rotation matrices", i);
      return 8;
    }
    bone->worlds = alloc_matrices(gsa->num_frames);
    if (! bone->worlds) {
      ERROR("out of memory for bone %u world rotation matrices", i);
      return 9;
    }
  }

  for (i = 0; i < gsa->num_frames; i++) {
    frame = gsa->frames[i];
    pos = get_file_position(gsa->fd);
    if (pos == -1) return 7;
    frame->offset = pos;
    if (parse_frame(gsa, i, frame, gsa->num_frames)) {
      ERROR("expected frame id %u but got null\n", i);
      return 12;
    }
  }

  pos = get_file_position(gsa->fd);
  if (fseek(gsa->fd, 0, SEEK_END)) {
    ERROR("can't seek to end of file to check if we've read it all");
    return 0;
  }
  if (get_file_position(gsa->fd) != pos) WARN("read last frame but not at end of file: offset 0x%04x, end: 0x%04x\n", pos, get_file_position(gsa->fd));
  return 0;
}

int parse_frame(struct gsm *gsa, unsigned short num, struct frame *frame, unsigned short total) {
  quaternion_t *orientation, *q, c, *tmp;
  struct bone *bone, *parent;
  int bone_id;
  unsigned short i;

  if (! gsa || ! gsa->fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  TRACE("reading frame %u offset relative to origin at 0x%04x\n", num, get_file_position(gsa->fd));
  /* co-ordinates of animation base wrt origin */
  if (read_vector(gsa->fd, &frame->position)) {
    ERROR("read failed parsing frame co-ords relative to OriginObject");
    return 2;
  }
  DEBUG("frame %u offset from origin: (% 0.4f, % 0.4f, % 0.4f)\n", num, frame->position.x, frame->position.y, frame->position.z);

  TRACE("reading frame %u animation base orientation at 0x%04x\n", num, get_file_position(gsa->fd));
  /* orientation of animation base */
  if (read_quaternion(gsa->fd, &frame->q)) {
    ERROR("read failed parsing frame %u rotation", num);
    return 5;
  }
  DEBUG("frame %u animation base: Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", num, frame->q.w, frame->q.x, frame->q.y, frame->q.z);

  TRACE("reading frame %u root bone position at 0x%04x\n", num, get_file_position(gsa->fd));
  /* position of root bone */
  if (read_vector(gsa->fd, &frame->root_position)) {
    ERROR("read failed parsing frame root bone co-ords");
    return 6;
  }
  DEBUG("frame %u root bone position: (% 0.4f, % 0.4f, % 0.4f)\n", num, frame->root_position.x, frame->root_position.y, frame->root_position.z);

  /* now the bones */
  if (! gsa->skeletons[0] || ! gsa->skeletons[0]->num_bones) {
    ERROR("no bones in gsa\n");
    return 9;
  }

  TRACE("reading frame %u orientations at 0x%04x\n", num, get_file_position(gsa->fd));
  for (i = 0; i < gsa->skeletons[0]->num_animated_bones; i++) {
    bone_id = get_bone_id_by_animated_bone_entry(gsa, i);
    if (bone_id < 0) {
      ERROR("can't find animated bone %u in animated bones list\n");
      return 10;
    }
    bone = gsa->skeletons[0]->bones[bone_id];
    orientation = bone->orientations[num];
    if (parse_orientation(gsa, i, orientation, num)) {
      ERROR("can't parse bone %u orientation\n", bone_id);
      return 11;
    }
    TRACE2("frame %d bone %u: (% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", num, bone_id, orientation->w, orientation->x, orientation->y, orientation->z);
  }

  /* some bones may not be animated - process after reading all orientations */
  if (calculate_animation_matrices(gsa, num)) {
    ERROR("can't calculate world orientation for frame %u\n", num);
    return 13;
  }

  if (find_events(gsa, frame, num)) {
    ERROR("error parsing frame %u events\n", num);
    return 14;
  }

  DEBUG("read frame %u (file pointer now at 0x%04x)\n", num, get_file_position(gsa->fd));

  return 0;
}

int find_events(struct gsm *gsa, struct frame *frame, unsigned short num) {
  unsigned short i, len;

  if (! gsa || ! gsa->fd) {
    ERROR("no file descriptor for GSA\n");
    return 1;
  }

  if (gsa->version >> 8 == 4 && gsa->version % 16 < 2) {
#ifdef YEAH_THIS_IS_SO_MUCH_EASIER_TO_READ_AND_BITSHIFTING_IS_SO_EXPENSIVE
    DEBUG("skipping events because this animation was exported with version %u.%u\n", (unsigned char) *((unsigned char *) &gsa->version + 1), (unsigned char) *((unsigned char *) &gsa->version));
#endif
    DEBUG("skipping events because this animation was exported with version %u.%u\n", gsa->version >> 8, gsa->version % 16);
    return 0;
  }

  if (read_short(gsa->fd, &frame->num_events)) {
    ERROR("can't read frame %u number of events", num);
    return 2;
  }
  if (! frame->num_events) {
    DEBUG("frame %u has no events\n", num);
    return 0;
  }
  gsa->skeletons[0]->num_events += frame->num_events;

  frame->events = alloc_events(frame->num_events);
  if (! frame->events) {
    ERROR("out of memory for %u events in frame %u\n", frame->num_events, num);
    return 3;
  }

  for (i = 0; i < frame->num_events; i++) {
    if (read_short(gsa->fd, &len)) {
      ERROR("can't read frame %u event %u name length", num, i);
      return 4;
    }
    frame->events[i] = alloc_event(len);
    if (! frame->events[i]) {
      ERROR("out of memory for frame %u event %u name\n", num, i);
      return 5;
    }
    if (read_arbitrary(gsa->fd, frame->events[i], len)) {
      ERROR("can't read frame %u event %u name", num, i);
      return 6;
    }
    DEBUG("frame %u event %u: %s (length %u)\n", num, i, frame->events[i], len);
  }

  return 0;
}

int write_animation_stuff(struct gsm *gsa) {
  unsigned short i;

  if (! gsa || ! gsa->out_fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  if (write_short(gsa->out_fd, &gsa->animation_unknown_before_framerate)) {
    ERROR("can't write unknown before framerate");
    return 2;
  }
  if (write_short(gsa->out_fd, &gsa->framerate)) {
    ERROR("can't write framerate");
    return 3;
  }
  if (write_short(gsa->out_fd, &gsa->animation_unknown_after_framerate)) {
    ERROR("can't write unknown after framerate");
    return 4;
  }

  return 0;
}

int write_frames(struct gsm *gsa) {
  struct frame *frame;
  unsigned short i, j, frames_written, frame_count, repeat_count;
  unsigned int old_pos, pos;
  int speedhack = 0;

  if (! gsa || ! gsa->out_fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  /* to slow animations down, we repeat a frame every other frame or so */
  frames_written = frame_count = 0;

  DEBUG("writing frames: %u\n", gsa->num_frames);
  old_pos = get_file_position(gsa->out_fd);
  /* number of frames */
  if (write_short(gsa->out_fd, &gsa->num_frames)) {
    ERROR("can't write number of frames");
    return 2;
  }
  /* shouldn't be possible anyway */
  if (! gsa->num_frames) return 0;

  if (gsa->option->frame_repeat_count > 1) {
    SAY("slowing animation down in %s\n", gsa->basename);
    speedhack = 1;
  }
  for (i = 0; i < gsa->num_frames; i++) {
    frame = gsa->frames[i];
    if (speedhack) {
      if (! (frame_count++ % gsa->option->frame_repeat_every)) {
        repeat_count = gsa->option->frame_repeat_count;
        DEBUG("will write frame %u %u times\n", i, repeat_count);
      }
    }
    else repeat_count = 1;
    /* write the frame this number of times */
    for (j = 0; j < repeat_count; j++) {
      if (write_frame(gsa, i, frame, gsa->num_frames)) {
        ERROR("can't write frame %u\n", i);
        return 3;
      }
      if (frames_written++ == 0xffff) {
        if (i < gsa->num_frames - 1) {
          ERROR("wrote too many frames, maximum possible %u\n", 0xffff);
          return 4;
        }
      }
    }
  }

  if (frames_written > gsa->num_frames) {
    pos = get_file_position(gsa->out_fd);
    if (fseek(gsa->out_fd, old_pos, SEEK_SET)) {
      ERROR("can't seek back to frame count to write real number of frames");
      return 5;
    }
    DEBUG("recording number of frames actually written: %u\n", frames_written);
    if (write_short(gsa->out_fd, &frames_written)) {
      ERROR("can't write real number of frames");
      return 6;
    }
    if (fseek(gsa->out_fd, pos, SEEK_SET)) {
      ERROR("can't seek back after frames");
      return 7;
    }
  }

  return 0;
}

int write_frame(struct gsm *gsa, unsigned short num, struct frame *frame, unsigned short total) {
  quaternion_t *orientation;
  unsigned short i;

  if (! gsa || ! gsa->out_fd) {
    ERROR("no file descriptor for gsa!\n");
    return 1;
  }

  DEBUG("writing frame %u at 0x%04x\n", num, get_file_position(gsa->out_fd));

  /* co-ordinates of animation base wrt origin */
  TRACE1("writing frame %u co-ordinates at 0x%04x\n", num, get_file_position(gsa->out_fd));
  if (write_vector(gsa->out_fd, &frame->position)) {
    ERROR("write failed on frame co-ords relative to OriginObject\n");
    return 2;
  }

  /* orientation of animation base */
  TRACE1("writing frame %u base orientation at 0x%04x\n", num, get_file_position(gsa->out_fd));
  if (write_quaternion(gsa->out_fd, &frame->q)) {
    ERROR("write failed on frame %u rotation\n", num);
    return 3;
  }

  /* position of root bone */
  TRACE1("writing frame %u root offset at 0x%04x\n", num, get_file_position(gsa->out_fd));
  if (write_vector(gsa->out_fd, &frame->root_position)) {
    ERROR("write failed on frame root bone co-ords\n");
    return 4;
  }

  /* now the bones */
  if (! gsa->skeletons[0] || ! gsa->skeletons[0]->num_bones) return 0;
  TRACE("writing frame %u orientations at 0x%04x\n", num, get_file_position(gsa->out_fd));
  for (i = 0; i < gsa->skeletons[0]->num_bones; i++) {
    if (! gsa->skeletons[0]->bones[i]->animated) continue;
    TRACE1("writing frame %u bone %u 0x%04x\n", num, i, get_file_position(gsa->out_fd));
    orientation = gsa->skeletons[0]->bones[i]->orientations[num];
    if (write_orientation(gsa, i, orientation, num)) {
      ERROR("can't write bone %u orientation\n", i);
      return 9;
    }
  }
  if (write_events(gsa, frame, num)) {
    ERROR("can't write frame %u events", num);
    return 10;
  }

  return 0;
}

int write_events(struct gsm *gsa, struct frame *frame, unsigned short num) {
  unsigned short i, len;

  if (! gsa || ! gsa->out_fd) {
    ERROR("no file descriptor for GSA\n");
    return 1;
  }

  TRACE1("writing frame %u events at 0x%04x\n", num, get_file_position(gsa->out_fd));
  if (write_short(gsa->out_fd, &frame->num_events)) {
    ERROR("can't write frame %u number of events", num);
    return 2;
  }
  if (! frame->num_events) return 0;

  for (i = 0; i < frame->num_events; i++) {
    len = strlen((const char *) frame->events[i]) + 1;
    if (write_short(gsa->out_fd, &len)) {
      ERROR("can't write frame %u event %u name length", num, i);
      return 3;
    }
    if (write_arbitrary(gsa->out_fd, frame->events[i], len)) {
      ERROR("can't write frame %u event %u name", num, i);
      return 4;
    }
  }

  return 0;
}
#if 0
/* bone's orientation is in frame list but we don't want it */
int delete_bone_from_frames(struct gsm *gsa, unsigned short bone_id) {
  if (! gsa || ! gsa->skeletons) {
    ERROR("no GSA\n");
    return 1;
  }
  if (! gsa->num_frames || ! gsa->frames) {
    ERROR("no frames\n");
    return 2;
  }

  DEBUG("removing bone %02u %s from %u frames\n", bone_id, get_biped_bone_name(gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES, bone_id), gsa->num_frames);

  return 0;
}
#endif

unsigned char **alloc_events(unsigned short n) {
  return (unsigned char **) alloc_lots(n, "event", sizeof(char *), 0);
}

unsigned char *alloc_event(unsigned short len) {
  return (unsigned char *) alloc_one("event", sizeof(char) * len);
}

void cleanup_events(unsigned short num_events, unsigned char **events) {
  free_structs("event", num_events, (void **) events);
}

struct event_set *alloc_event_set() {
  return (struct event_set *) alloc_one("event_set", sizeof(struct event_set));
}

int add_to_event_set(struct event_set **event_sets, char *frame_arg, char *event) {
  unsigned short len;
  struct event_set *last, *tmp;
  char *bogus;

  tmp = alloc_event_set();
  if (! tmp) {
    ERROR("out of memory for event_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  tmp->frame_id = (unsigned short) strtol(frame_arg, &bogus, 0);
  if (*bogus) {
    ERROR("bogus frame ID in event set\n");
    return 2;
  }

  if (event) {
    tmp->event = alloc_name(event, &len);
    if (! len) {
      ERROR("out of memory for event_set event");
      return 3;
    }
  }

  TRACE("tmp->frame_id: %u, tmp->event: %s\n", tmp->frame_id, tmp->event);

  last = (struct event_set *) gll_last((struct gll *) *event_sets);
  TRACE("last: %p\n", last);
  if (last) last->next = tmp;
  else *event_sets = tmp;
  /*gll_debug((struct gll *) *event_sets);*/

  return 0;
}

int set_events(struct gsm *gsa) {
  struct event_set *event_set;
  struct frame *frame;
  unsigned char *event, **new_events;
  unsigned short expected, replaced;
  unsigned short i, num_events, len;
  char *bogus;

  if (! gsa || ! gsa->num_frames) {
    ERROR("no frames in gsa\n");
    return -1;
  }

  if (! gsa->type & GSM_TYPE_ANIMATION) {
    ERROR("not an animation\n");
    return -1;
  }

  if (! gsa->option->event_sets) {
    ERROR("no set parameters\n");
    return -2;
  }

  expected = replaced = 0;
  GLL_LOOP(event_set, gsa->option->event_sets) {
    expected++;

    /* check valid frame */
    if (event_set->frame_id > gsa->num_frames) {
      error_dbg("frame id %u out of range\n", event_set->frame_id);
      continue;
    }
    frame = gsa->frames[event_set->frame_id];

    /* null event => --clear-events was called */
    if (! event_set->event) {
      replaced++;
      SAY("clearing events from frame %u\n", event_set->frame_id);
      cleanup_events(frame->num_events, frame->events);
      frame->num_events = 0;
      continue;
    }

    /* make the new event */
    event = alloc_name((char *) event_set->event, &len);
    if (! len) {
      ERROR("out of memory for new event");
      return -3;
    }

    DEBUG("adding event \"%s\" to frame %u\n", event, event_set->frame_id);

    /* if we have already been called, append us to the map list */
    if (frame->changed) {
      num_events = frame->num_events + 1;
      new_events = alloc_events(num_events);
      for (i = 0; i < frame->num_events; i++) {
        new_events[i] = alloc_name((char *) frame->events[i], &len);
        if (! len) {
          ERROR("out of memory setting new event name %u\n", i);
          return -4;
        }
      }
    }
    else {
      num_events = 1;
      new_events = alloc_events(num_events);
      frame->num_events = 0;
    }
    new_events[frame->num_events] = event;
    cleanup_events(frame->num_events, frame->events);
    frame->events = new_events;
    frame->num_events = num_events;
    frame->changed = 1;

    /* cool, it worked */
    replaced++;
  }


  if (replaced == 1) SAY("set one event\n");
  else if (replaced) SAY("set %u events\n", replaced);
  else if (replaced != expected) error_dbg("no events set, expected to set %u\n", expected);
  else error_dbg("no events set\n");

  return replaced;
}

char *get_events(struct frame *frame) {
  unsigned short i;
  int len;
  char *ret, *s;
  unsigned char *event;

  if (! frame) {
    ERROR("no frame\n");
    return "bogus";
  }

  if (! frame->num_events) {
    ERROR("no events in frame\n");
    return "no events";
  }

  len = 0;
  for (i = 0; i < frame->num_events; i++) len += strlen((const char *) frame->events[i]);

  ret = (char *) malloc(sizeof(char) * (len + 3 * (frame->num_events - 1) + 1));
  if (! ret) {
    ERROR("out of memory for event names\n");
    return "unknown";
  }

  s = ret;
  TRACE("getting events: %u\n", frame->num_events);
  for (i = 0; i < frame->num_events; i++) {
    event = frame->events[i];
    TRACE1("writing event %u \"%s\" to %p\n", i, event, ret);
    len = snprintf(s, strlen((const char *) event) + 1, "%s", event);
    if (len == -1) {
      ERROR("snprintf() failed for event %u\n", i);
      free(ret);
      return "unknown";
    }
    s += len;
    if (i < frame->num_events - 1) {
      TRACE1("writing comma %u to %p\n", i, s);
      len = snprintf(s, 3, ", ");
      if (len == -1) {
        ERROR("snprintf() failed for comma %u\n", i);
        free(ret);
        return "unknown";
      }
    }
    TRACE2("string ret at %p is \"%s\"\n", ret, ret);
    s += len;
  }

  return ret;
}

struct rotation_set *alloc_rotation_set() {
  return (struct rotation_set *) alloc_one("rotation_set", sizeof(struct rotation_set));
}

void cleanup_rotation_sets(struct rotation_set *rotation_sets) {
  struct rotation_set *tmp;
  GLL_LOOP(tmp, (struct gll *) rotation_sets) {
    TRACE2("freeing rotation_set %p bone\n", tmp);
    if (tmp->bone_arg) free(tmp->bone_arg);
  }
  TRACE2("freeing rotation_sets at %p\n", rotation_sets);
  gll_free((struct gll *) rotation_sets);
}

int add_to_rotation_set(struct rotation_set **rotation_sets, char *bone_arg, char *axis, char *angle_arg, char *first_frame_arg, char *last_frame_arg) {
  struct rotation_set *last, *tmp;
  char *bogus;
  int bone_id, num;
  float angle;
  unsigned short len;
  matrix_t *m;

  tmp = alloc_rotation_set();
  if (! tmp) {
    ERROR("out of memory for rotation_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  /* get bone to rotate */
  bone_id = (int) strtol(bone_arg, &bogus, 0);
  if (*bogus) {
    /* assume bone name */
    tmp->bone_arg = alloc_name(bone_arg, &len);
    if (! len) {
      ERROR("out of memory for rotation_set->bone_arg");
      return 2;
    }
    TRACE("tmp->bone_arg: %p\n", tmp->bone_arg);
  }
  else {
    /* could be a direct bone ID */
    if (bone_id < 0) {
      ERROR("negative bone ID\n");
      return 4;
    }
    if (bone_id > 0xffff) {
      ERROR("bone ID out of range\n");
      return 5;
    }
    tmp->bone_id = (unsigned short) bone_id;
    TRACE("tmp->bone_id: %u\n", tmp->bone_id);
  }

  /* get rotation angle */
  angle = (float) strtod(angle_arg, &bogus);
  if (*bogus || angle < -180.0 || angle > 180.0) {
    ERROR("bogus rotation angle: try -180 to 180\n");
    return 6;
  }
  if (! angle) {
    ERROR("rotating through 0 degrees is kind of pointless\n");
    return 7;
  }

  /* get rotation axis */
  if (str_equiv(axis, "x") || str_equiv(axis, "x-axis") || str_equiv(axis, "x axis")) m = make_rotation_matrix_in_x(degrees_to_radians(angle));
  else if (str_equiv(axis, "y") || str_equiv(axis, "y-axis") || str_equiv(axis, "y axis")) m = make_rotation_matrix_in_y(degrees_to_radians(angle));
  else if (str_equiv(axis, "z") || str_equiv(axis, "z-axis") || str_equiv(axis, "z axis")) m = make_rotation_matrix_in_z(degrees_to_radians(angle));
  else {
    ERROR("bogus rotation axis: try X, Y or Z\n");
    return 8;
  }
  if (! m) {
    ERROR("can't convert rotation angles to matrix\n");
    return 9;
  }

  /* convert to matrix */
  copy_matrix(&tmp->rotation, m);
  free(m);

  /* get first frame */
  num = (int) strtol(first_frame_arg, &bogus, 0);
  if (*bogus) {
    /* "first" is OK */
    if (str_equiv(first_frame_arg, "first")) num = 0;
    else {
      ERROR("bogus value for first frame: try a number\n");
      return 10;
    }
  }
  else {
    if (num < 0) {
      ERROR("negative first frame number\n");
      return 11;
    }
    if (num > 0xffff) {
      ERROR("first frame number out of range\n");
      return 12;
    }
  }
  tmp->first_frame = (unsigned short) num;
  TRACE("tmp->first_frame: %u\n", tmp->first_frame);

  /* get last frame */
  num = (int) strtol(last_frame_arg, &bogus, 0);
  if (*bogus) {
    /* "last" is OK */
    if (str_equiv(last_frame_arg, "last")) num = -1;
    else {
      ERROR("bogus value for last frame: try a number or the word \"last\"\n");
      return 13;
    }
  }
  else {
    if (num > 0xffff) {
      ERROR("last frame number out of range\n");
      return 14;
    }
    if (num < tmp->first_frame) {
      ERROR("last frame < first frame\n");
      return 15;
    }
  }
  tmp->last_frame = num;
  TRACE("tmp->last_frame: %u\n", tmp->last_frame);

  last  = (struct rotation_set *) gll_last((struct gll *) *rotation_sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *rotation_sets = tmp;
  /*gll_debug((struct gll *) *rotation_sets);*/

  return 0;
}

/* not a general function - this handles the --rotate-bone argument */
int rotate_bones(struct gsm *gsa) {
  struct rotation_set *rotation_set, *rotation_sets;
  int errors, ret;

  if (! gsa) {
    ERROR("no GSA!\n");
    return 1;
  }
  if (! gsa->is_animation) {
    error_dbg("can only rotate bones in animations\n");
    return 2;
  }

  rotation_sets = gsa->option->rotation_sets;
  if (! rotation_sets) {
    ERROR("no rotation sets\n");
    return 3;
  }

  SAY("rotating bones\n");
  errors = 0;
  GLL_LOOP(rotation_set, rotation_sets) {
    ret = rotate_bone(gsa, rotation_set);
    if (ret) {
      ERROR("rotate_bone() returned %d\n", ret);
      errors++;
    }
  }

  if (errors) {
    ERROR("%d error%s occurred rotating bones\n", errors, (errors == 1) ? "" : "s");
    errors += 3;
  }
  return errors;
}

int rotate_bone(struct gsm *gsa, struct rotation_set *rotation_set) {
  unsigned short i;
  int ret;
  struct bone *bone;
  matrix_t *m;

  if (! gsa || ! rotation_set) {
    ERROR("bogus arguments!\n");
    return 1;
  }
  if (! gsa->is_animation) {
    error_dbg("can only rotate bones in animations\n");
    return 2;
  }

  /* bone may have been specified by name */
  if (rotation_set->bone_arg) {
    ret = get_bone_by_name(gsa, rotation_set->bone_arg);
    if (ret < 0) {
      ERROR("can't rotate a non-existent bone!\n");
      return 3;
    }
    rotation_set->bone_id = (unsigned short) ret;
  }
  bone = gsa->skeletons[0]->bones[rotation_set->bone_id];
  if (! bone) {
    ERROR("null bone!\n");
    return 4;
  }

  /* last frame may have been given as "last" */
  if (rotation_set->last_frame == -1) {
    INFO("last frame in animation is %u\n", gsa->num_frames);
    rotation_set->last_frame = gsa->num_frames - 1;
  }

  /* check last frame in range */
  if (rotation_set->last_frame >= gsa->num_frames) {
    ERROR("can't rotate bone up to frame %u because there are only %u frames\n", rotation_set->last_frame, gsa->num_frames);
    return 5;
  }

  for (i = rotation_set->first_frame; i <= rotation_set->last_frame; i++) {
    /* rotate bone in local co-ordinates for this frame */
    m = multiply_matrices(&rotation_set->rotation, bone->locals[i]);
    if (! m) {
      ERROR("can't multiply matrices\n");
      return 6;
    }
    copy_matrix(bone->locals[i], m);
    free(m);

    if (matrix_to_quaternion(bone->orientations[i], bone->locals[i])) {
      ERROR("can't convert new rotation matrix to quaternion\n");
      return 7;
    }

    /* update skeleton hierarchy's rotations */
    if (bone->links_to == BONE_FIRST_LINK) {
      /* copy local rotation as world and calculate children's rotations */
      copy_matrix(bone->worlds[i], bone->locals[i]);
      ret = rotation_set->bone_id;
    }
    else {
      /* calculate hierarchy from parent (cheat) */
      ret = gsa->skeletons[0]->bones[rotation_set->bone_id]->links_to;
    }
    if (calculate_world_for_all_children(gsa, ret, i)) {
      ERROR("can't recalculate world rotations\n");
      return 8;
    }
  }

  /* mark bone as animated */
  bone->animated = 1;

  return 0;
}

/*
  move down through the skeleton and work out the world orientation of each 
  bone starting from the parent - all this happens in frame num
*/
int calculate_world_for_all_children(struct gsm *gsa, unsigned short bone_id, unsigned short num) {
  unsigned short i;
  matrix_t *m;
  struct bone *bone, *child;

  if (! gsa) {
    ERROR("no GSA!\n");
    return 1;
  }
  if (num >= gsa->num_frames) {
    ERROR("frame ID out of range\n");
    return 2;
  }
  if (bone_id >= gsa->skeletons[0]->num_bones) {
    ERROR("bone ID out of range\n");
    return 3;
  }
  bone = gsa->skeletons[0]->bones[bone_id];

  /*
    fundamental assumption:
    a bone cannot be a child of a bone with a higher bone ID
  */
  TRACE("calculating world orientations for children of bone %u in frame %u\n", bone_id, num);
  for (i = bone_id + 1; i < gsa->skeletons[0]->num_bones; i++) {
    child = gsa->skeletons[0]->bones[i];
    if (child->links_to != bone_id) continue;
    TRACE1("calculating world orientation for bone %u in frame %u\n", i, num);
    m = multiply_matrices(bone->worlds[num], child->locals[num]);
    if (! m) {
      ERROR("can't calculate world orientation for %s\n", child->name);
      return 4;
    }
    copy_matrix(child->worlds[num], m);
    free(m);
  }

  return 0;
}

/*
  having read in all local rotations for the frame, let's calculate world
  the major concern here is that some bones aren't animated so we need to 
  explicitly calculate their positions before plugging their world into 
  the child's local
*/
int calculate_animation_matrices(struct gsm *gsa, unsigned short num) {
  unsigned short i;
  struct bone *bone, *parent;
  matrix_t *m;

  if (! gsa || ! gsa->skeletons || ! gsa->skeletons[0]) {
    ERROR("no GSA!\n");
    return 1;
  }
  if (num >= gsa->num_frames) {
    ERROR("frame %u out of range\n", num);
    return 2;
  }

  for (i = 0; i < gsa->skeletons[0]->num_bones; i++) {
    bone = gsa->skeletons[0]->bones[i];

    /* first determine local rotation matrix for the frame */
    if (bone->animated) {
      if (quaternion_to_matrix(bone->locals[num], bone->orientations[num])) {
        ERROR("can't convert frame %u orientation to local rotation matrix for bone %u\n", num, i);
        return 3;
      }
    }
    else {
      TRACE("bone %u isn't animated - copying default pose frame orientations\n", i);
      copy_matrix(bone->locals[num], &bone->local);
    }

    if (set_matrix_translation(bone->locals[num], &bone->position)) {
      ERROR("can't set translation part of local matrix for bone %u\n", i);
      return 4;
    }

    if (bone->links_to == BONE_FIRST_LINK) {
      /* this is the root bone so we can't get parent's world */
      copy_matrix(bone->worlds[num], bone->locals[num]);
      continue;
    }

    parent = gsa->skeletons[0]->bones[i - 1];
    m = multiply_matrices(parent->worlds[num], bone->locals[num]);
    if (! m) {
      ERROR("can't convert bone %u co-ordinate system in frame %u\n", i, num);
      return 5;
    }
    copy_matrix(bone->worlds[num], m);
    free(m);

    /* FIXME: calculate world position */
  }

  return 0;
}

int describe_frames(struct gsm *gsa) {
  unsigned short i, j;
  struct bone *bone;
  struct frame *frame;
  char *tmp;
  quaternion_t *q;

  if (! gsa) {
    ERROR("no GSA!\n");
    return 1;
  }

  for (i = 0; i < gsa->num_frames; i++) {
    frame = gsa->frames[i];
    if (gsm_verbose && ! i) SAY("\n");
    SAY("Frame %02u at 0x%04x", i, gsa->frames[i]->offset);
    if (frame->num_events) {
      tmp = get_events(frame);
      SAY(": %s", tmp);
      free(tmp);
    }
    SAY("\n");
    if (gsm_verbose) {
      SAY("  Frame position: (% 0.4f, % 0.4f, % 0.4f)\n", frame->position.x, frame->position.y, frame->position.z);
      SAY("  Frame orientation: Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", frame->q.w, frame->q.x, frame->q.y, frame->q.z);
      SAY("  Root bone position: (% 0.4f, % 0.4f, % 0.4f)\n", frame->root_position.x, frame->root_position.y, frame->root_position.z);
      if (gsm_verbose > 1) {
        for (j = 0; j < gsa->skeletons[0]->num_bones; j++) {
          bone = gsa->skeletons[0]->bones[j];
          if (! bone->animated) continue;
          q = bone->orientations[i];
          SAY("  Bone %02u orientation: Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", j, q->w, q->x, q->y, q->z/*, quaternion_magnitude(q)*/);
          if (gsm_verbose > 2) {
            SAY("  Local   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R0], bone->local[R3], bone->local[R6], bone->local[TX]);
            SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R1], bone->local[R4], bone->local[R7], bone->local[TY]);
            SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R2], bone->local[R5], bone->local[R8], bone->local[TZ]);

            SAY("\n");

            SAY("  World   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R0], bone->world[R3], bone->world[R6], bone->world[TX]);
            SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R1], bone->world[R4], bone->world[R7], bone->world[TY]);
            SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R2], bone->world[R5], bone->world[R8], bone->world[TZ]);
          }
        }
      }
    }
  }

  return 0;
}
